# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django.utils import timezone

from cronman.models import CronTask
from cronman.taxonomies import CronTaskStatus
from cronman.tests.base import BaseCronTestCase


class CronTaskTestCase(BaseCronTestCase):
    """Tests for CronTask model"""

    def test_job_spec_no_params(self):
        """Test for CronTask.job_spec method - no params"""
        cron_task = CronTask.objects.run_now("Sleep")[0]
        self.assertEqual(
            cron_task.job_spec(), "Sleep:task_id={}".format(cron_task.pk)
        )

    def test_job_spec_with_params(self):
        """Test for CronTask.job_spec method - with params"""
        cron_task = CronTask.objects.run_now("Sleep", params="seconds=42")[0]
        self.assertEqual(
            cron_task.job_spec(),
            "Sleep:seconds=42,task_id={}".format(cron_task.pk),
        )

    def test_waiting(self):
        """Test for CronTask model instance with WAITING status"""
        cron_task = CronTask.objects.run_now("Sleep")[0]
        self.assertEqual(cron_task.status, CronTaskStatus.WAITING)

        self.assertTrue(cron_task.is_pending())
        self.assertFalse(cron_task.is_started())
        self.assertFalse(cron_task.is_failed())
        self.assertFalse(cron_task.is_finished())

        self.assertEqual(list(CronTask.objects.pending()), [cron_task])
        self.assertEqual(list(CronTask.objects.started()), [])
        self.assertEqual(list(CronTask.objects.failed()), [])
        self.assertEqual(list(CronTask.objects.finished()), [])

    def test_queued(self):
        """Test for CronTask model instance with QUEUED status"""
        cron_task = CronTask.objects.run_now("Sleep")[0]
        cron_task.mark_as_queued()
        self.assertEqual(cron_task.status, CronTaskStatus.QUEUED)

        self.assertTrue(cron_task.is_pending())
        self.assertFalse(cron_task.is_started())
        self.assertFalse(cron_task.is_failed())
        self.assertFalse(cron_task.is_finished())

        self.assertEqual(list(CronTask.objects.pending()), [cron_task])
        self.assertEqual(list(CronTask.objects.started()), [])
        self.assertEqual(list(CronTask.objects.failed()), [])
        self.assertEqual(list(CronTask.objects.finished()), [])

    def test_started(self):
        """Test for CronTask model instance with STARTED status"""
        pid = 1001
        now = timezone.now()
        cron_task = CronTask.objects.run_now("Sleep")[0]
        cron_task.mark_as_started(pid, date_time=now)
        self.assertEqual(cron_task.status, CronTaskStatus.STARTED)
        self.assertEqual(cron_task.pid, pid)
        self.assertEqual(cron_task.started_at, now)

        self.assertFalse(cron_task.is_pending())
        self.assertTrue(cron_task.is_started())
        self.assertFalse(cron_task.is_failed())
        self.assertFalse(cron_task.is_finished())

        self.assertEqual(list(CronTask.objects.pending()), [])
        self.assertEqual(list(CronTask.objects.started()), [cron_task])
        self.assertEqual(list(CronTask.objects.failed()), [])
        self.assertEqual(list(CronTask.objects.finished()), [])

    def test_failed(self):
        """Test for CronTask model instance with FAILED status"""
        cron_task = CronTask.objects.run_now("Sleep")[0]
        cron_task.mark_as_failed()
        self.assertEqual(cron_task.status, CronTaskStatus.FAILED)

        self.assertFalse(cron_task.is_pending())
        self.assertFalse(cron_task.is_started())
        self.assertTrue(cron_task.is_failed())
        self.assertFalse(cron_task.is_finished())

        self.assertEqual(list(CronTask.objects.pending()), [])
        self.assertEqual(list(CronTask.objects.started()), [])
        self.assertEqual(list(CronTask.objects.failed()), [cron_task])
        self.assertEqual(list(CronTask.objects.finished()), [])

    def test_finished(self):
        """Test for CronTask model instance with FINISHED status"""
        now = timezone.now()
        cron_task = CronTask.objects.run_now("Sleep")[0]
        cron_task.mark_as_finished(date_time=now)
        self.assertEqual(cron_task.status, CronTaskStatus.FINISHED)
        self.assertEqual(cron_task.finished_at, now)

        self.assertFalse(cron_task.is_pending())
        self.assertFalse(cron_task.is_started())
        self.assertFalse(cron_task.is_failed())
        self.assertTrue(cron_task.is_finished())

        self.assertEqual(list(CronTask.objects.pending()), [])
        self.assertEqual(list(CronTask.objects.started()), [])
        self.assertEqual(list(CronTask.objects.failed()), [])
        self.assertEqual(list(CronTask.objects.finished()), [cron_task])



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import signal

import mock

from cronman.tests.base import BaseCronTestCase
from cronman.worker.signal_notifier import SignalNotifier


class SignalNotifierTestCase(BaseCronTestCase):
    """Tests for SignalNotifier class"""

    def test_context_manager_handlers_assignment(self):
        """Test for context manager: signal handlers (dis)assignment"""
        # `signal.default_int_handler` may be overwritten by unittests:
        default_int_handler = signal.getsignal(signal.SIGINT)
        default_term_handler = signal.getsignal(signal.SIGTERM)
        with SignalNotifier("TestCronJob") as signal_notifier:
            self.assertEqual(
                signal.getsignal(signal.SIGINT),
                signal_notifier.notify_and_quit,
            )
            self.assertEqual(
                signal.getsignal(signal.SIGTERM),
                signal_notifier.notify_and_quit,
            )
        self.assertEqual(signal.getsignal(signal.SIGINT), default_int_handler)
        self.assertEqual(
            signal.getsignal(signal.SIGTERM), default_term_handler
        )

    @mock.patch("cronman.worker.signal_notifier.sys.exit")
    def test_handle_sigint(self, mock_exit):
        """Test for `notify_and_quit`, case: SIGINT received"""
        signal_notifier = SignalNotifier("TestCronJob")
        signal_notifier.logger = mock.MagicMock()
        signal_notifier.slack = mock.MagicMock()

        signal_notifier.notify_and_quit(signal.SIGINT, None)

        mock_exit.assert_called_once_with(signal.SIGINT)
        signal_notifier.logger.warning.assert_has_calls(
            [mock.call('Cron job "TestCronJob" killed by SIGINT.')]
        )
        signal_notifier.slack.post.assert_called_once_with(
            'Cron job "TestCronJob" killed by SIGINT.'
        )

    @mock.patch("cronman.worker.signal_notifier.sys.exit")
    def test_handle_sigterm(self, mock_exit):
        """Test for `notify_and_quit`, case: SIGTERM received"""
        signal_notifier = SignalNotifier("TestCronJob")
        signal_notifier.logger = mock.MagicMock()
        signal_notifier.slack = mock.MagicMock()

        signal_notifier.notify_and_quit(signal.SIGTERM, None)

        mock_exit.assert_called_once_with(signal.SIGTERM)
        signal_notifier.logger.warning.assert_has_calls(
            [mock.call('Cron job "TestCronJob" killed by SIGTERM.')]
        )
        signal_notifier.slack.post.assert_called_once_with(
            'Cron job "TestCronJob" killed by SIGTERM.'
        )



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from cronman.tests.base import TEMP_FILE

CRON_JOBS = (
    ("*/2 * * * *", "Sleep:seconds=1,path={}".format(TEMP_FILE)),
    ("*/2 * * * *", "Sleep:seconds=2"),
)

ALLOWED_CRON_TASKS = ("Sleep", "ParamsLockedSleep", "ClassLockedSleep")



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import logging
import time
from collections import OrderedDict

from cronman.exceptions import PIDAccessError
from cronman.taxonomies import JobSpecStatus, PIDStatus
from cronman.utils import format_exception, parse_job_spec
from cronman.worker.worker_file import CronWorkerJobSpecFile, CronWorkerPIDFile

logger = logging.getLogger("cronman.command.cron_worker")


class BaseCronWorkerList(object):
    """Listing Cron Worker processes - base class"""

    file_class = None  # must be set in subclass

    def __init__(
        self,
        data_dir,
        job_spec=None,
        name=None,
        args=None,
        kwargs=None,
        pid=None,
    ):
        if pid:
            file_by_pid = self.file_class.by_pid(data_dir, pid)
            files = [file_by_pid] if file_by_pid else []
        else:
            if job_spec:
                name, args, kwargs = parse_job_spec(job_spec)
            # Files sorted by name:
            files = sorted(
                self.file_class.all(data_dir, name, args, kwargs),
                key=lambda pf: pf.name,
            )
        self.data_dir = data_dir
        self.files = files
        self.logger = logger


class CronWorkerPIDList(BaseCronWorkerList):
    """Listing and killing Cron Worker processes through PID files"""

    file_class = CronWorkerPIDFile
    wait_to_kill = 7

    def _iter_status_items(self):
        """Iterator over status information dicts extracted from files"""
        for pid_file in self.files:
            pid = pid_file.pid
            process_exists = pid_file.process.exists()
            if process_exists is None:
                error = PIDAccessError(
                    "{} No access to PID {}!".format(pid_file.name, pid)
                )
                self.logger.warning(format_exception(error))
                continue
            if process_exists:
                status = PIDStatus.ALIVE
            else:
                status = PIDStatus.DEAD
            item = OrderedDict()
            item["name"] = pid_file.name
            item["status"] = status
            item["pid"] = pid
            item["_pid_file"] = pid_file
            yield item

    def status(self):
        """Retrieves status information about listed PID files"""
        items = []
        totals = OrderedDict()
        totals["TOTAL"] = 0
        totals[PIDStatus.ALIVE] = 0
        totals[PIDStatus.DEAD] = 0
        for item in self._iter_status_items():
            totals[item["status"]] += 1
            totals["TOTAL"] += 1
            items.append(item)
        return items, totals

    def clean(self):
        """Removes dead PID files"""
        items = []
        totals = OrderedDict()
        totals["TOTAL"] = 0
        for item in self._iter_status_items():
            if item["status"] == PIDStatus.DEAD:
                item["_pid_file"].delete()
                item["status"] = PIDStatus.DELETED
            else:
                continue
            totals["TOTAL"] += 1
            items.append(item)
        return items, totals

    def kill(self):
        """Kills Cron Worker processes associated with listed PID files"""
        items = []
        totals = OrderedDict()
        totals["TOTAL"] = 0
        totals[PIDStatus.DEAD] = 0
        totals[PIDStatus.TERMED] = 0
        totals[PIDStatus.KILLED] = 0
        # Round 1: TERMINATE
        at_least_one_termed = False
        for item in self._iter_status_items():
            if item["status"] == PIDStatus.ALIVE:
                item["_pid_file"].process.terminate()
                item["status"] = PIDStatus.TERMED
                at_least_one_termed = True
            items.append(item)
        # Round 2: wait and KILL:
        if at_least_one_termed:
            wait = True
            for item in items:
                if (
                    item["status"] == PIDStatus.TERMED
                    and item["_pid_file"].process.alive()
                ):
                    # Sleep only once, before first kill:
                    if wait:
                        time.sleep(self.wait_to_kill)
                        wait = False
                        # Skip if process finished while we were sleeping:
                        if not item["_pid_file"].process.alive():
                            continue
                    item["_pid_file"].process.kill()
                    item["status"] = PIDStatus.KILLED
        # Update totals:
        for item in items:
            totals[item["status"]] += 1
            totals["TOTAL"] += 1
        return items, totals


class CronWorkerJobSpecList(BaseCronWorkerList):
    """Listing and resuming Cron Worker processes through JobSpec files"""

    file_class = CronWorkerJobSpecFile

    def _iter_status_items(self):
        """Iterator over status information dicts extracted from files"""
        for job_spec_file in self.files:
            pid_file = job_spec_file.pid_file
            if pid_file.exists():
                process_exists = pid_file.process.exists()
                if process_exists is None:
                    error = PIDAccessError(
                        "{} No access to PID {}!".format(
                            pid_file.name, pid_file.pid
                        )
                    )
                    self.logger.warning(format_exception(error))
                    continue
                if process_exists:
                    status = JobSpecStatus.ACTIVE
                else:
                    status = JobSpecStatus.STALLED
            else:
                status = JobSpecStatus.STALLED
            item = OrderedDict()
            item["name"] = job_spec_file.name
            item["status"] = status
            item["job_spec"] = job_spec_file.job_spec
            item["_job_spec_file"] = job_spec_file
            yield item

    def status(self):
        """Retrieves status information about listed JobSpec files"""
        items = []
        totals = OrderedDict()
        totals["TOTAL"] = 0
        totals[JobSpecStatus.ACTIVE] = 0
        totals[JobSpecStatus.STALLED] = 0
        for item in self._iter_status_items():
            totals[item["status"]] += 1
            totals["TOTAL"] += 1
            items.append(item)
        return items, totals

    def clean(self):
        """Removes stalled JobSpec files"""
        items = []
        totals = OrderedDict()
        totals["TOTAL"] = 0
        for item in self._iter_status_items():
            if item["status"] == JobSpecStatus.STALLED:
                item["_job_spec_file"].delete()
                item["status"] = JobSpecStatus.DELETED
            else:
                continue
            totals["TOTAL"] += 1
            items.append(item)
        return items, totals

    def resume(self):
        """Starts Cron Worker process for each stalled JobSpec file (resume)"""
        items = []
        totals = OrderedDict()
        totals["TOTAL"] = 0
        for item in self._iter_status_items():
            if item["status"] == JobSpecStatus.STALLED:
                pid = item["_job_spec_file"].resume()  # file gets deleted
                if pid is None:
                    continue
                item["status"] = JobSpecStatus.RESUMED
            else:
                continue
            totals["TOTAL"] += 1
            items.append(item)
        return items, totals



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django.contrib import admin
from django.utils.module_loading import import_string

from cronman.config import app_settings
from cronman.forms import CronTaskAdminForm
from cronman.models import CronTask

admin_site_path = app_settings.CRONMAN_ADMIN_SITE

admin_site = import_string(admin_site_path) if admin_site_path else None


class CronTaskAdmin(admin.ModelAdmin):
    """Admin interface options for CronTask model"""

    form = CronTaskAdminForm
    list_display = (
        "job_spec",
        "start_at",
        "status",
        "user",
        "created_at",
        "started_at",
        "finished_at",
    )
    list_filter = ("status", "cron_job")
    max_num = 0

    def save_model(self, request, obj, form, change):
        """Assign current user to the task"""
        obj.user = request.user
        return super(CronTaskAdmin, self).save_model(
            request, obj, form, change
        )

    def get_fields(self, request, obj=None):
        """Fields: all fields on "edition", editable fields on addition"""
        if obj:
            fields = [
                "cron_job",
                "params",
                "start_at",
                "status",
                "user",
                "created_at",
                "started_at",
                "finished_at",
            ]
        else:
            fields = ["cron_job", "params", "start_at"]
        return fields

    def get_readonly_fields(self, request, obj=None):
        """Readonly Fields: all on "edition", none on addition"""
        if obj:
            readonly_fields = self.get_fields(request, obj)
        else:
            readonly_fields = []
        return readonly_fields


if admin_site:
    admin_site.register(CronTask, CronTaskAdmin)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import errno
import platform

import mock

from cronman.spawner import CronSpawner
from cronman.tests.base import (
    TEST_CRONMAN_DATA_DIR,
    BaseCronTestCase,
    expected_worker_env,
    mock_environ,
    override_cron_settings,
)

SYSTEM_NAME = platform.node()


class CronSpawnerTestCase(BaseCronTestCase):
    """Tests for CronSpawner class"""

    @override_cron_settings()
    def test_get_worker_env(self):
        """Test for CronSpawner.get_worker_env method"""
        self.assertDictContainsSubset(
            {
                "CRONMAN_DATA_DIR": TEST_CRONMAN_DATA_DIR,
                "CRONMAN_CRONITOR_ENABLED": "0",
            },
            CronSpawner().get_worker_env(),
        )

    @mock.patch("cronman.spawner.spawn")
    @mock.patch("cronman.spawner.time.sleep")
    def test_start_ok(self, mock_sleep, mock_spawn):
        """Test for CronSpawner.start_worker method - OK"""
        spawner = CronSpawner()
        spawner.slack = mock.MagicMock()
        spawner.logger = mock.MagicMock()

        spawner.start_worker("Sleep:seconds=1")

        mock_sleep.assert_not_called()
        self.assertEqual(mock_spawn.call_count, 1)
        self.assertFalse(spawner.memory_error_occurred)

    @mock.patch(
        "cronman.spawner.spawn",
        side_effect=OSError(errno.ENOMEM, "Cannot allocate memory"),
    )
    @mock.patch("cronman.spawner.time.sleep")
    def test_start_worker_out_of_memory(self, mock_sleep, mock_spawn):
        """Test for CronSpawner.start_worker method - OOM error case"""
        spawner = CronSpawner()
        spawner.slack = mock.MagicMock()
        spawner.logger = mock.MagicMock()

        spawner.start_worker("Sleep:seconds=1")

        spawner.slack.post.assert_has_calls(
            [
                mock.call(
                    "[{}] "
                    "RuntimeError: "
                    "Unable to start a worker process "
                    "for Sleep:seconds=1 due to Out-Of-Memory error.".format(
                        SYSTEM_NAME
                    )
                )
            ]
        )
        spawner.logger.warning.assert_has_calls(
            [
                mock.call(
                    "RuntimeError: "
                    "Unable to start a worker process "
                    "for Sleep:seconds=1 due to Out-Of-Memory error."
                )
            ]
        )
        mock_sleep.assert_called_once_with(spawner.wait_for_memory)
        self.assertEqual(mock_spawn.call_count, 2)
        self.assertTrue(spawner.memory_error_occurred)

    @mock.patch(
        "cronman.spawner.spawn",
        side_effect=OSError(errno.ENOMEM, "Cannot allocate memory"),
    )
    @mock.patch("cronman.spawner.time.sleep")
    def test_start_worker_out_of_memory_again(self, mock_sleep, mock_spawn):
        """Test for CronSpawner.start_worker method - OOM error case (again).
        """
        spawner = CronSpawner()
        spawner.slack = mock.MagicMock()
        spawner.logger = mock.MagicMock()
        spawner.memory_error_occurred = True

        spawner.start_worker("Sleep:seconds=1")

        spawner.slack.post.assert_has_calls(
            [
                mock.call(
                    "[{}] "
                    "RuntimeError: "
                    "Unable to start a worker process "
                    "for Sleep:seconds=1 due to Out-Of-Memory error.".format(
                        SYSTEM_NAME
                    )
                )
            ]
        )
        spawner.logger.warning.assert_has_calls(
            [
                mock.call(
                    "RuntimeError: "
                    "Unable to start a worker process "
                    "for Sleep:seconds=1 due to Out-Of-Memory error."
                )
            ]
        )
        mock_sleep.assert_not_called()
        self.assertEqual(mock_spawn.call_count, 1)
        self.assertTrue(spawner.memory_error_occurred)

    @mock.patch(
        "cronman.spawner.spawn",
        side_effect=OSError(errno.EPERM, "Permission Denied"),
    )
    @mock.patch("cronman.spawner.time.sleep")
    def test_start_with_os_error(self, mock_sleep, mock_spawn):
        """Test for CronSpawner.start_worker method - other OSError case"""
        spawner = CronSpawner()
        spawner.slack = mock.MagicMock()
        spawner.logger = mock.MagicMock()

        with self.assertRaises(OSError):
            spawner.start_worker("Sleep:seconds=1")

        mock_sleep.assert_not_called()
        self.assertEqual(mock_spawn.call_count, 1)
        self.assertFalse(spawner.memory_error_occurred)

    @mock.patch("cronman.spawner.spawn", side_effect=RuntimeError)
    @mock.patch("cronman.spawner.time.sleep")
    def test_start_with_other_error(self, mock_sleep, mock_spawn):
        """Test for CronSpawner.start_worker method - other exception type
        case"""
        spawner = CronSpawner()
        spawner.slack = mock.MagicMock()
        spawner.logger = mock.MagicMock()

        with self.assertRaises(RuntimeError):
            spawner.start_worker("Sleep:seconds=1")

        mock_sleep.assert_not_called()
        self.assertEqual(mock_spawn.call_count, 1)
        self.assertFalse(spawner.memory_error_occurred)

    @override_cron_settings()
    @mock.patch("cronman.spawner.spawn")
    @mock.patch("cronman.spawner.os.environ.copy", mock_environ)
    @mock.patch(
        "cronman.spawner.sys.argv", ["manage.py", "cron_scheduler", "run"]
    )
    @mock.patch("cronman.spawner.sys.executable", "/bin/python")
    def test_start_worker_os_environ_passed(self, mock_spawn):
        """Test that CronSpawner.start_worker passes converted environment
        variables to child processes.
        """
        spawner = CronSpawner()
        spawner.start_worker("Sleep:seconds=1,path=/tmp/sleep.txt")
        mock_spawn.assert_called_once_with(
            "/bin/python",
            "manage.py",
            "cron_worker",
            "run",
            "Sleep:seconds=1,path=/tmp/sleep.txt",
            env=expected_worker_env(),
        )

    @override_cron_settings(CRONMAN_RAVEN_CMD=None)
    @mock.patch("cronman.spawner.spawn")
    @mock.patch(
        "cronman.spawner.sys.argv", ["manage.py", "cron_scheduler", "run"]
    )
    @mock.patch("cronman.spawner.sys.executable", "/bin/python")
    def test_start_worker_no_raven_cmd_no_settings(self, mock_spawn):
        """Test for CronSpawner.start_worker method
         - CRONMAN_RAVEN_CMD defined, no --settings passed
        """
        spawner = CronSpawner()
        spawner.start_worker("Sleep:seconds=10")
        mock_spawn.assert_called_once_with(
            "/bin/python",
            "manage.py",
            "cron_worker",
            "run",
            "Sleep:seconds=10",
            env=spawner.get_worker_env(),
        )

    @override_cron_settings(CRONMAN_RAVEN_CMD=None)
    @mock.patch("cronman.spawner.spawn")
    @mock.patch(
        "cronman.spawner.sys.argv",
        ["manage.py", "cron_scheduler", "run", "--settings=test"],
    )
    @mock.patch("cronman.spawner.sys.executable", "/bin/python")
    def test_start_worker_no_raven_cmd_with_settings(self, mock_spawn):
        """Test for CronSpawner.start_worker method
         - CRONMAN_RAVEN_CMD undefined, --settings passed
        """
        spawner = CronSpawner()
        spawner.start_worker("Sleep:seconds=10")
        mock_spawn.assert_called_once_with(
            "/bin/python",
            "manage.py",
            "cron_worker",
            "run",
            "Sleep:seconds=10",
            "--settings=test",
            env=spawner.get_worker_env(),
        )

    @override_cron_settings(CRONMAN_RAVEN_CMD=None)
    @mock.patch("cronman.spawner.spawn")
    @mock.patch(
        "cronman.spawner.sys.argv", ["manage.py", "cron_scheduler", "run"]
    )
    @mock.patch("cronman.spawner.sys.executable", "/bin/python")
    def test_start_worker_no_raven_cmd_no_settings_quoted(self, mock_spawn):
        """Test for CronSpawner.start_worker method
         - CRONMAN_RAVEN_CMD defined, no --settings passed, quoted param
        """
        spawner = CronSpawner()
        spawner.start_worker('Sleep:seconds=10,quoted="This is a test"')
        mock_spawn.assert_called_once_with(
            "/bin/python",
            "manage.py",
            "cron_worker",
            "run",
            'Sleep:seconds=10,quoted="This is a test"',
            env=spawner.get_worker_env(),
        )

    @override_cron_settings(CRONMAN_RAVEN_CMD="/usr/bin/raven-cmd")
    @mock.patch("cronman.spawner.spawn")
    @mock.patch(
        "cronman.spawner.sys.argv", ["manage.py", "cron_scheduler", "run"]
    )
    @mock.patch("cronman.spawner.sys.executable", "/bin/python")
    def test_start_worker_with_raven_cmd_no_settings(self, mock_spawn):
        """Test for CronSpawner.start_worker method
         - CRONMAN_RAVEN_CMD defined, no --settings passed
        """
        spawner = CronSpawner()
        spawner.start_worker("Sleep:seconds=10")
        mock_spawn.assert_called_once_with(
            "/usr/bin/raven-cmd",
            "-c",
            "/bin/python manage.py cron_worker run Sleep:seconds=10",
            env=spawner.get_worker_env(),
        )

    @override_cron_settings(CRONMAN_RAVEN_CMD="/usr/bin/raven-cmd")
    @mock.patch("cronman.spawner.spawn")
    @mock.patch(
        "cronman.spawner.sys.argv",
        ["manage.py", "cron_scheduler", "run", "--settings=test"],
    )
    @mock.patch("cronman.spawner.sys.executable", "/bin/python")
    def test_start_worker_with_raven_cmd_with_settings(self, mock_spawn):
        """Test for CronSpawner.start_worker method
         - CRONMAN_RAVEN_CMD defined, --settings passed
        """
        spawner = CronSpawner()
        spawner.start_worker("Sleep:seconds=10")
        mock_spawn.assert_called_once_with(
            "/usr/bin/raven-cmd",
            "-c",
            "/bin/python manage.py cron_worker run Sleep:seconds=10 "
            "--settings=test",
            env=spawner.get_worker_env(),
        )

    @override_cron_settings(CRONMAN_RAVEN_CMD="/usr/bin/raven-cmd")
    @mock.patch("cronman.spawner.spawn")
    @mock.patch(
        "cronman.spawner.sys.argv", ["manage.py", "cron_scheduler", "run"]
    )
    @mock.patch("cronman.spawner.sys.executable", "/bin/python")
    def test_start_worker_with_raven_cmd_no_settings_quoted(self, mock_spawn):
        """Test for CronSpawner.start_worker method
         - CRONMAN_RAVEN_CMD defined, no --settings passed, quoted param
        """
        spawner = CronSpawner()
        spawner.start_worker('Sleep:seconds=10,quoted="This is a test"')
        mock_spawn.assert_called_once_with(
            "/usr/bin/raven-cmd",
            "-c",
            "/bin/python manage.py cron_worker run "
            "\"Sleep:seconds=10,quoted='This is a test'\"",
            env=spawner.get_worker_env(),
        )

    # Starting workers with CPU/IO priority

    @override_cron_settings(CRONMAN_RAVEN_CMD=None)
    @mock.patch("cronman.spawner.spawn")
    @mock.patch(
        "cronman.spawner.sys.argv", ["manage.py", "cron_scheduler", "run"]
    )
    @mock.patch("cronman.spawner.sys.executable", "/bin/python")
    def test_start_worker_nice_ionice_no_raven_cmd_no_settings(
        self, mock_spawn
    ):
        """Test for CronSpawner.start_worker method
         - CRONMAN_RAVEN_CMD defined, no --settings passed, nice, ionice
        """
        spawner = CronSpawner()
        spawner.start_worker("LowestCPUIOSleep:seconds=10")
        mock_spawn.assert_called_once_with(
            "nice",
            "-n",
            "19",
            "ionice",
            "-c",
            "2",
            "-n",
            "7",
            "/bin/python",
            "manage.py",
            "cron_worker",
            "run",
            "LowestCPUIOSleep:seconds=10",
            env=spawner.get_worker_env(),
        )

    @override_cron_settings(CRONMAN_RAVEN_CMD="/usr/bin/raven-cmd")
    @mock.patch("cronman.spawner.spawn")
    @mock.patch(
        "cronman.spawner.sys.argv", ["manage.py", "cron_scheduler", "run"]
    )
    @mock.patch("cronman.spawner.sys.executable", "/bin/python")
    def test_start_worker_ionice_with_raven_cmd_no_settings_quoted(
        self, mock_spawn
    ):
        """Test for CronSpawner.start_worker method
         - CRONMAN_RAVEN_CMD defined, no --settings passed, quoted param, ionice
        """
        spawner = CronSpawner()
        spawner.start_worker('IdleIOSleep:seconds=10,quoted="This is a test"')
        mock_spawn.assert_called_once_with(
            "/usr/bin/raven-cmd",
            "-c",
            "ionice -c 3 /bin/python manage.py cron_worker run "
            "\"IdleIOSleep:seconds=10,quoted='This is a test'\"",
            env=spawner.get_worker_env(),
        )

    @override_cron_settings(CRONMAN_RAVEN_CMD="/usr/bin/raven-cmd")
    @mock.patch("cronman.spawner.spawn")
    @mock.patch(
        "cronman.spawner.sys.argv", ["manage.py", "cron_scheduler", "run"]
    )
    @mock.patch("cronman.spawner.sys.executable", "/bin/python")
    def test_start_worker_nice_with_raven_cmd_no_settings(self, mock_spawn):
        """Test for CronSpawner.start_worker method
         - CRONMAN_RAVEN_CMD defined, no --settings passed, nice
        """
        spawner = CronSpawner()
        spawner.start_worker("LowCPUSleep:seconds=10")
        mock_spawn.assert_called_once_with(
            "/usr/bin/raven-cmd",
            "-c",
            "nice -n 10 /bin/python manage.py cron_worker run "
            "LowCPUSleep:seconds=10",
            env=spawner.get_worker_env(),
        )



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import logging
import signal
import sys

from cronman.monitor import Slack

logger = logging.getLogger("cronman.command.cron_worker")


class SignalNotifier(object):
    """Class responsible for sending notifications when SIGINT / SIGTERM
    is received.
    """

    signal_names = ((signal.SIGINT, "SIGINT"), (signal.SIGTERM, "SIGTERM"))

    def __init__(self, job_name):
        self.job_name = job_name
        self.logger = logger
        self.slack = Slack()
        self.previous_int_handler = signal.getsignal(signal.SIGINT)
        self.previous_term_handler = signal.getsignal(signal.SIGTERM)

    def notify_and_quit(self, signum, frame):
        """Handles a signal by sending notification and quitting"""
        message = 'Cron job "{}" killed by {}.'.format(
            self.job_name, dict(self.signal_names).get(signum, signum)
        )
        self.logger.warning(message)
        self.slack.post(message)
        sys.exit(signum)

    def capture(self):
        """Assigns handlers to signals"""
        self.previous_int_handler = signal.signal(
            signal.SIGINT, self.notify_and_quit
        )
        self.previous_term_handler = signal.signal(
            signal.SIGTERM, self.notify_and_quit
        )

    def reset(self):
        """Restores previous signal handlers"""
        signal.signal(signal.SIGINT, self.previous_int_handler)
        signal.signal(signal.SIGTERM, self.previous_term_handler)

    def __enter__(self):
        self.capture()
        return self

    def __exit__(self, *args, **kwargs):
        self.reset()



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from collections import OrderedDict

from cronman.job import cron_job_registry
from cronman.utils import function_signature


class CronJobClassList(object):
    """Listing cron job classes"""

    def __init__(self, name=None, cron_job_class=None):
        if name or cron_job_class:
            if not name:
                name = cron_job_class.__name__
            elif not cron_job_class:
                cron_job_class = cron_job_registry.get(name)
            items = [(name, cron_job_class)]
        else:
            items = sorted(
                cron_job_registry.items(),
                # Sort by module path, name:
                key=lambda t: (t[1].__module__, t[0]),
            )
        self.items = items

    def _iter_info_items(self):
        """Iterator over information dicts for available cron job classes."""
        for name, cron_job_class in self.items:
            item = OrderedDict()
            item["name"] = name
            item["class"] = "{}.{}".format(
                cron_job_class.__module__, cron_job_class.__name__
            )
            item["_cron_job_class"] = cron_job_class
            yield item

    def summary(self):
        """Retrieves brief information about available cron job classes"""
        items = list(self._iter_info_items())
        totals = {"TOTAL": len(items)}
        return items, totals

    def details(self):
        """Retrieves more information about available cron job classes"""
        items = []
        totals = {"TOTAL": 0}
        for item in self._iter_info_items():
            cron_job_class = item["_cron_job_class"]
            params = function_signature(cron_job_class.run)
            if params:
                item["params"] = params
            description = "\n".join(
                line.strip()
                for line in (cron_job_class.__doc__ or "").split("\n")
            )
            if description:
                item["description"] = "\n" + description
            totals["TOTAL"] += 1
            items.append(item)
        return items, totals



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django.utils.module_loading import import_string

from cronman.config import app_settings
from cronman.exceptions import MissingDependency


def get_strict_redis(host=None, port=None, db=None):
    """Retrieves Redis client object (StrictRedis) using constructor function
    defined in settings (`CRONMAN_REDIS_CONSTRUCTOR`).
    """
    _get_strict_redis = import_string(app_settings.CRONMAN_REDIS_CONSTRUCTOR)
    return _get_strict_redis(host=host, port=port, db=db)


def get_strict_redis_default(host=None, port=None, db=None):
    """Retrieve Redis client object (StrictRedis) - default implementation."""
    try:
        from redis import StrictRedis
    except ImportError:
        raise MissingDependency(
            "Unable to import redis. "
            "CronRemoteManager requires this dependency."
        )

    host = host or app_settings.CRONMAN_REDIS_HOST
    port = port or app_settings.CRONMAN_REDIS_PORT
    db = db if db is not None else app_settings.CRONMAN_REDIS_DB

    return StrictRedis(host=host, port=port, db=db)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from cronman.config import app_settings
from cronman.job import BaseCronJob
from cronman.models import CronTask
from cronman.taxonomies import PIDStatus
from cronman.utils import cron_jobs_module_config
from cronman.worker import CronWorker


class CleanCronTasks(BaseCronJob):
    """Changes status of dead CronTasks from STARTED to FAILED."""

    cronitor_id = app_settings.CRONMAN_CLEAN_CRON_TASKS_CRONITOR_ID

    def __init__(self, logger=None):
        super(CleanCronTasks, self).__init__(logger=logger)
        self.cron_worker = CronWorker()
        self.cron_worker.logger = self.logger

    def run(self):
        """Main logic"""
        num_failed = 0
        cron_tasks = self.get_started_cron_tasks()
        if cron_tasks:
            active_pids = self.get_active_pids()
            for cron_task in cron_tasks:
                if cron_task.pid not in active_pids:
                    cron_task.mark_as_failed()
                    num_failed += 1
        if num_failed:
            status_message = "{} CronTask(s) marked as failed.".format(
                num_failed
            )
        else:
            status_message = "No CronTasks marked as failed."
        self.logger.info(status_message)

    def get_started_cron_tasks(self):
        """Retrieves started CronTasks"""
        allowed_tasks = cron_jobs_module_config(
            "ALLOWED_CRON_TASKS", default=()
        )
        return CronTask.objects.started().filter(cron_job__in=allowed_tasks)

    def get_active_pids(self):
        """Retrieves list of all active cron worker PIDs"""
        current_cron_jobs = self.cron_worker.get_worker_pid_list().status()[0]
        return [
            item["pid"]
            for item in current_cron_jobs
            if item["status"] == PIDStatus.ALIVE
        ]



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import datetime

from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.six import python_2_unicode_compatible

from cronman.taxonomies import CronTaskStatus


class CronTaskQuerySet(models.QuerySet):
    """QuerySet for CronTask model"""

    def pending(self):
        """CronTasks with status suitable for start - WAITING or QUEUED"""
        return self.filter(
            status__in=(CronTaskStatus.WAITING, CronTaskStatus.QUEUED)
        )

    def started(self):
        """CronTasks with status STARTED (running)"""
        return self.filter(status=CronTaskStatus.STARTED)

    def failed(self):
        """CronTasks with status FAILED"""
        return self.filter(status=CronTaskStatus.FAILED)

    def finished(self):
        """CronTasks with status FINISHED (success)"""
        return self.filter(status=CronTaskStatus.FINISHED)


# pylint: disable=no-member
class CronTaskManager(models.Manager.from_queryset(CronTaskQuerySet)):
    """Manager for CronTask model"""

    def run_now(self, cron_job, params="", user=None, now=None):
        """Requests a cron job to run now by creating a CronTask instance.
        Prevents duplicated objects from being created.
        """
        now = now or timezone.now()
        tolerance = datetime.timedelta(minutes=4)
        cron_task = self.filter(
            cron_job=cron_job,
            params=params,
            start_at__gt=now - tolerance,
            start_at__lt=now + tolerance,
        ).first()
        if cron_task:
            created = False
        else:
            cron_task = self.create(
                cron_job=cron_job, params=params, user=user, start_at=now
            )
            created = True
        return cron_task, created


@python_2_unicode_compatible
class CronTask(models.Model):
    """Cron Task: a cron job execution request"""

    TASK_ID_PARAM = "task_id"

    cron_job = models.CharField(max_length=255)
    params = models.TextField(blank=True)
    start_at = models.DateTimeField(blank=True, default=timezone.now)

    status = models.CharField(
        max_length=16,
        choices=CronTaskStatus.CHOICES,
        default=CronTaskStatus.WAITING,
    )
    pid = models.PositiveIntegerField(null=True, blank=True)

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    created_at = models.DateTimeField(null=True, auto_now_add=True)
    started_at = models.DateTimeField(null=True, blank=True)
    finished_at = models.DateTimeField(null=True, blank=True)

    objects = CronTaskManager()

    def job_spec(self):
        """Cron Job spec for this task"""
        cron_job = self.cron_job
        params = self.params.strip()
        params += "{}{}={}".format(
            "," if params else "", self.TASK_ID_PARAM, self.pk
        )
        return "{}:{}".format(cron_job, params)

    def __str__(self):
        return self.job_spec()

    # Status check methods:

    def is_pending(self):
        """Check if this CronTask is suitable to launch - WAITING or QUEUED"""
        return self.status in (CronTaskStatus.WAITING, CronTaskStatus.QUEUED)

    def is_started(self):
        """Check if this CronTask has status STARTED (running)"""
        return self.status == CronTaskStatus.STARTED

    def is_failed(self):
        """Check if this CronTask has status FAILED"""
        return self.status == CronTaskStatus.FAILED

    def is_finished(self):
        """Check if this CronTask has status FINISHED (success)"""
        return self.status == CronTaskStatus.FINISHED

    # Status change methods:

    def mark_as_queued(self):
        """Marks this CronTask as QUEUED"""
        self.status = CronTaskStatus.QUEUED
        self.save()

    def mark_as_started(self, pid, date_time=None):
        """Marks this CronTask as QUEUED - retrieved by worker process."""
        self.status = CronTaskStatus.STARTED
        self.pid = pid
        self.started_at = date_time or timezone.now()
        self.save()

    def mark_as_failed(self):
        """Marks this CronTask as FAILED"""
        self.status = CronTaskStatus.FAILED
        self.save()

    def mark_as_finished(self, date_time=None):
        """Marks this CronTask as FINISHED successfully"""
        self.status = CronTaskStatus.FINISHED
        self.finished_at = date_time or timezone.now()
        self.save()



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django.db import connections
from django.utils import timezone
from django.utils.functional import cached_property

from cronman.config import app_settings
from cronman.job import BaseCronJob
from cronman.models import CronTask
from cronman.spawner import CronSpawner
from cronman.utils import cron_jobs_module_config


class RunCronTasks(BaseCronJob):
    """Starts worker processes for cron jobs requested to run in Admin
    via CronTask model.
    """

    lock_ignore_errors = True
    cronitor_id = app_settings.CRONMAN_RUN_CRON_TASKS_CRONITOR_ID

    @cached_property
    def cron_spawner(self):
        """Cron Spawner instance"""
        return CronSpawner(logger=self.logger)

    def run(self):
        """Main logic"""
        cron_tasks = self.get_pending_cron_tasks()
        num_cron_tasks = len(cron_tasks)
        num_started = 0
        for i, cron_task in enumerate(cron_tasks, 1):
            self.logger.info(
                "Starting worker for CronTask {} ({}/{})".format(
                    cron_task, i, num_cron_tasks
                )
            )
            pid = self.start_cron_task(cron_task)
            if pid is not None:
                num_started += 1
        if num_started:
            status_message = "Started {} CronTask(s).".format(num_started)
        else:
            status_message = "No CronTasks started."
        self.logger.info(status_message)

    def get_pending_cron_tasks(self):
        """Retrieve pending CronTasks"""
        allowed_tasks = cron_jobs_module_config(
            "ALLOWED_CRON_TASKS", default=()
        )
        cron_tasks = list(
            CronTask.objects.pending()
            .filter(start_at__lte=timezone.now())
            .filter(cron_job__in=allowed_tasks)
        )
        connections.close_all()  # close db connections
        return cron_tasks

    def start_cron_task(self, cron_task):
        """Starts worker for given CronTask"""
        return self.cron_spawner.start_worker(cron_task.job_spec())



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import platform

import mock

from cronman.exceptions import CronWorkerInvalidParams
from cronman.models import CronTask
from cronman.taxonomies import CronTaskStatus
from cronman.tests.base import (
    BaseCronTestCase,
    create_pid_file,
    override_cron_settings,
    patch_kill,
    patch_ps,
)
from cronman.worker import CronWorker

SYSTEM_NAME = platform.node()


class CronWorkerTestCase(BaseCronTestCase):
    """Tests for CronWorker class"""

    # Main methods

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.Sleep.run")
    def test_run_without_params(self, mock_run):
        """Test for CronWorker.run method - run cron job without params"""
        worker = CronWorker()
        output = worker.run("Sleep")
        self.assertIn("OK: Processed Sleep", output)
        mock_run.assert_called_once_with()

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.Sleep.run")
    def test_run_with_params(self, mock_run):
        """Test for CronWorker.run method - run cron job with params"""
        worker = CronWorker()
        output = worker.run("Sleep:42,path=/tmp/test")
        self.assertIn("OK: Processed Sleep:42,path=/tmp/test", output)
        mock_run.assert_called_once_with("42", path="/tmp/test")

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.Sleep.run")
    def test_run_non_existing_cron_job(self, mock_run):
        """Test for CronWorker.run method - attempt to run non existing
        cron job
        """
        worker = CronWorker()
        with self.assertRaisesMessage(
            CronWorkerInvalidParams,
            "CronJobNotRegistered: {!r}".format("NoSuchWorker"),
        ):
            worker.run("NoSuchWorker")
        mock_run.assert_not_called()

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.Sleep.run")
    def test_run_invalid_params(self, mock_run):
        """Test for CronWorker.run method - attempt to run cron job with
        invalid params
        """
        worker = CronWorker()
        with self.assertRaisesMessage(
            CronWorkerInvalidParams,
            "ValueError: In chars 9-10 `5`: "
            "Positional argument after named arguments.",
        ):
            worker.run("Sleep:2,test=3,5")  # positional arg. after named one
        mock_run.assert_not_called()

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.Sleep.run", side_effect=ValueError)
    def test_run_cron_job_error(self, mock_run):
        """Test for CronWorker.run method - exception raised while processing
        """
        worker = CronWorker()
        output = worker.run("Sleep")
        self.assertIn("FAIL: Processed Sleep", output)
        mock_run.assert_called_once_with()

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.ClassLockedSleep.run")
    def test_run_cron_job_while_class_lock_enabled(self, mock_run):
        """Test for CronWorker.run method -
        attempt to run while class-based lock enabled.
        """
        worker = CronWorker()
        worker.slack = mock.MagicMock()
        create_pid_file("ClassLockedSleep")
        output = worker.run("ClassLockedSleep")
        self.assertIn(
            "CronWorkerLocked: "
            'Unable to start "ClassLockedSleep", '
            "because similar process is already running (PID file exists).",
            output,
        )
        mock_run.assert_not_called()
        worker.slack.post.assert_called_once_with(
            (
                "[{}] "
                "CronWorkerLocked: "
                'Unable to start "ClassLockedSleep", '
                "because similar process is already running (PID file exists)."
            ).format(SYSTEM_NAME)
        )

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.ParamsLockedSleep.run")
    def test_run_cron_job_while_params_lock_enabled(self, mock_run):
        """Test for CronWorker.run method -
        attempt to run while params-based lock enabled.
        """
        worker = CronWorker()
        worker.slack = mock.MagicMock()
        create_pid_file("ParamsLockedSleep")
        output = worker.run("ParamsLockedSleep")
        self.assertIn(
            "CronWorkerLocked: "
            'Unable to start "ParamsLockedSleep", '
            "because similar process is already running (PID file exists).",
            output,
        )
        mock_run.assert_not_called()
        worker.slack.post.assert_called_once_with(
            (
                "[{}] "
                "CronWorkerLocked: "
                'Unable to start "ParamsLockedSleep", '
                "because similar process is already running (PID file exists)."
            ).format(SYSTEM_NAME)
        )

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.IgnoreLockErrorsSleep.run")
    def test_run_cron_job_with_lock_ignore_errors(self, mock_run):
        """Test for CronWorker.run method -
        attempt to run while params-based lock enabled and `lock_ignore_errors`
        option is set.
        """
        worker = CronWorker()
        worker.slack = mock.MagicMock()
        create_pid_file("IgnoreLockErrorsSleep")
        output = worker.run("IgnoreLockErrorsSleep")
        self.assertIn(
            "CronWorkerLocked: "
            'Unable to start "IgnoreLockErrorsSleep", '
            "because similar process is already running (PID file exists).",
            output,
        )
        mock_run.assert_not_called()
        worker.slack.post.assert_not_called()

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.SlackNotifyDoneSleep.run")
    def test_run_cron_job_with_slack_notify_done(self, mock_run):
        """Test for CronWorker.run method - run cron job with
        `slack_notify_done` option set.
        """
        worker = CronWorker()
        worker.slack = mock.MagicMock()
        output = worker.run("SlackNotifyDoneSleep")
        self.assertIn("OK: Processed SlackNotifyDoneSleep", output)
        mock_run.assert_called_once_with()
        worker.slack.post.assert_called_once_with(
            'Cron job "SlackNotifyDoneSleep" is done.'
        )

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.Sleep.run")
    def test_run_cron_job_without_slack_notify_done(self, mock_run):
        """Test for CronWorker.run method - run cron job without
        `slack_notify_done` option set.
        """
        worker = CronWorker()
        worker.slack = mock.MagicMock()
        output = worker.run("Sleep")
        self.assertIn("OK: Processed Sleep", output)
        worker.slack.post.assert_not_called()

    # Tests for Cron Worker + Cron Tasks:

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.Sleep.run")
    def test_run_missing_task(self, mock_run):
        """Test for CronWorker.run method - run DELETED CronTask"""
        cron_task = CronTask.objects.run_now("Sleep")[0]
        job_spec = cron_task.job_spec()
        cron_task.delete()
        worker = CronWorker()
        output = worker.run(job_spec)

        mock_run.assert_called_once_with()
        self.assertIn("OK:", output)  # Cron job passes anyway

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.Sleep.run")
    def test_run_task_without_params(self, mock_run):
        """Test for CronWorker.run method - run CronTask without params"""
        cron_task = CronTask.objects.run_now("Sleep")[0]
        worker = CronWorker()
        worker.run(cron_task.job_spec())

        mock_run.assert_called_once_with()
        cron_task.refresh_from_db()
        self.assertEqual(cron_task.status, CronTaskStatus.FINISHED)

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.Sleep.run")
    def test_run_task_with_params(self, mock_run):
        """Test for CronWorker.run method - run CronTask with params"""
        cron_task = CronTask.objects.run_now(
            "Sleep", params="42,path=/tmp/test"
        )[0]
        worker = CronWorker()
        worker.run(cron_task.job_spec())

        mock_run.assert_called_once_with("42", path="/tmp/test")
        cron_task.refresh_from_db()
        self.assertEqual(cron_task.status, CronTaskStatus.FINISHED)

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.Sleep.run", side_effect=ValueError)
    def test_run_task_error(self, mock_run):
        """Test for CronWorker.run method - exception raised while processing
        CronTask.
        """
        cron_task = CronTask.objects.run_now("Sleep")[0]
        worker = CronWorker()
        worker.run(cron_task.job_spec())

        mock_run.assert_called_once_with()
        cron_task.refresh_from_db()
        self.assertEqual(cron_task.status, CronTaskStatus.FAILED)

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.ClassLockedSleep.run")
    def test_run_task_while_class_lock_enabled(self, mock_run):
        """Test for CronWorker.run method -
        attempt to run CronTask while class-based lock enabled.
        """
        cron_task = CronTask.objects.run_now("ClassLockedSleep")[0]
        worker = CronWorker()
        worker.slack = mock.MagicMock()
        create_pid_file(cron_task.cron_job)
        worker.run(cron_task.job_spec())

        mock_run.assert_not_called()
        cron_task.refresh_from_db()
        self.assertEqual(cron_task.status, CronTaskStatus.QUEUED)

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.ClassLockedSleep.run")
    def test_run_task_with_invalid_status(self, mock_run):
        """Test for CronWorker.run method -
        attempt to run CronTask with invalid status.
        """
        cron_task = CronTask.objects.run_now("ClassLockedSleep")[0]
        cron_task.mark_as_failed()

        worker = CronWorker()
        worker.slack = mock.MagicMock()
        worker.logger = mock.MagicMock()
        output = worker.run(cron_task.job_spec())

        message = (
            "CronTaskInvalidStatus: "
            'Unable to start "ClassLockedSleep:task_id={}", '
            "because associated CronTask "
            'has invalid status "Failed".'.format(cron_task.pk)
        )
        self.assertIn(message, output)
        worker.logger.warning.assert_called_once_with(message)
        worker.slack.post.assert_called_once_with(
            "[{}] {}".format(SYSTEM_NAME, message)
        )
        mock_run.assert_not_called()

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.ClassLockedSleep.run")
    def test_run_killed_task(self, mock_run):
        """Test for CronWorker.run method - resuming killed CronTask."""
        pid = 1001
        cron_task = CronTask.objects.run_now("ClassLockedSleep")[0]
        cron_task.mark_as_started(pid)
        with patch_ps():
            with patch_kill():
                worker = CronWorker()
                worker.slack = mock.MagicMock()
                worker.logger = mock.MagicMock()
                output = worker.run(cron_task.job_spec())

        self.assertIn("OK:", output)
        worker.logger.info.assert_has_calls(
            [
                mock.call(
                    'Starting "ClassLockedSleep:task_id={}" '
                    "for killed CronTask.".format(cron_task.pk)
                )
            ]
        )
        mock_run.assert_called_once_with()

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.sleep.ClassLockedSleep.run")
    def test_run_task_already_started(self, mock_run):
        """Test for CronWorker.run method -
        attempt to run already started CronTask.
        """
        pid = 1001
        cron_task = CronTask.objects.run_now("ClassLockedSleep")[0]
        cron_task.mark_as_started(pid)
        with patch_ps(active_pids=[pid]):
            with patch_kill(active_pids=[pid]):
                worker = CronWorker()
                worker.slack = mock.MagicMock()
                worker.logger = mock.MagicMock()
                output = worker.run(cron_task.job_spec())

        message = (
            "CronTaskInvalidStatus: "
            'Unable to start "ClassLockedSleep:task_id={}", '
            "because associated CronTask "
            'has invalid status "Started".'.format(cron_task.pk)
        )
        self.assertIn(message, output)
        worker.logger.warning.assert_called_once_with(message)
        worker.slack.post.assert_called_once_with(
            "[{}] {}".format(SYSTEM_NAME, message)
        )
        mock_run.assert_not_called()



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import datetime
import inspect
import logging
import os
import pipes
import re
import subprocess
import sys
from importlib import import_module

from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.six import text_type
# pylint: disable=E0401, E0611
from django.utils.six.moves import range

from dateutil.parser import parse as dateutil_parse

from cronman.config import app_settings

MYPY = False
if MYPY:
    # pylint: disable=unused-import
    from typing import Union, Text, TypeVar

    T = TypeVar("T")

logger = logging.getLogger("cronman.command")


PARAM_PATTERN = re.compile(
    r"(\s*(?P<key>[\w\d_\-]*)\s*=\s*)?((?P<value>(\"[^\"]*\"|'[^']*'|[^,]*)),?)"
)


def format_exception(exception=None):
    """Convert exception into string for logging purposes."""
    exception = exception if exception is not None else sys.exc_info()[1]
    return (
        "{}: {}".format(type(exception).__name__, force_text(exception))
        if exception is not None
        else ""
    )


def config(name):
    """Retrieves config value from environment or app settings"""
    value = os.environ.get(name)
    if value is None:
        value = getattr(app_settings, name)
    return value


def cron_jobs_module_config(name, default=None):
    """Retrieves value as attribute of CRONMAN_JOBS_MODULE."""
    cron_jobs_module = config("CRONMAN_JOBS_MODULE")
    if cron_jobs_module is None:
        return default
    # AttributeError or ImportError should be loud
    value = getattr(import_module(cron_jobs_module), name)
    return default if value is None else value


def chunks(iterable, n):
    """Yield successive n-sized chunks from iterable"""
    for i in range(0, len(iterable), n):
        yield iterable[i : i + n]


def function_signature(func):
    """Formats function parameters"""
    arg_spec = inspect.getargspec(func)
    defaults = arg_spec.defaults or ()
    num_args = len(arg_spec.args)
    args = []
    for i, name in enumerate(arg_spec.args):
        # NOTE: In Python 3, unbound methods are functions so we can't use
        # `inspect.ismethod`. Instead, we use naming convention to eliminate
        # first parameter when it's a reference to current object or class.
        if i == 0 and name in ("self", "cls", "mcs"):
            continue
        try:
            default = defaults[-num_args + i]
        except IndexError:
            args.append(name)
        else:
            args.append("{}={!r}".format(name, default))
    if arg_spec.varargs:
        args.append("*{}".format(arg_spec.varargs))
    if arg_spec.keywords:
        args.append("**{}".format(arg_spec.keywords))
    return ", ".join(args)


def _parse_params_error(match_obj, message):
    """Helper to raise error in parse_params"""
    return ValueError(
        "In chars {}-{} `{}`: {}".format(
            match_obj.start(), match_obj.end(), match_obj.group(0), message
        )
    )


def parse_params(params_string):
    """Converts string of params into pair: (positional args, named args)"""
    args = []
    kwargs = {}
    match_objects = list(PARAM_PATTERN.finditer(params_string))
    for match_obj in match_objects[:-1]:  # skip empty end match
        group_dict = match_obj.groupdict()

        # Clean value:
        value = group_dict["value"].strip()
        if not value:
            raise _parse_params_error(
                match_obj, 'Implicit empty value. Use explicit `""` instead.'
            )
        if value.count('"') % 2 or value.count("'") % 2:
            raise _parse_params_error(match_obj, "Unbalanced parentheses.")
        if value[0] == value[-1] and value[0] in ("'", '"'):
            value = value[1:-1]

        key = group_dict["key"]
        if key is None:
            # Positional argument:
            if kwargs:
                raise _parse_params_error(
                    match_obj, "Positional argument after named arguments."
                )
            args.append(value)
        else:
            # Named argument:
            if not key:
                raise _parse_params_error(match_obj, "Empty named argument.")
            if key in kwargs:
                raise _parse_params_error(
                    match_obj, "Duplicated named argument."
                )
            kwargs[key] = value

    return args, kwargs


def parse_job_spec(job_spec):
    """Converts a string into pair: CronJob name, params dict"""
    try:
        name, params_string = job_spec.split(":", 1)
    except ValueError:
        name, params_string = job_spec, ""
    if params_string:
        args, kwargs = parse_params(params_string)
    else:
        args, kwargs = [], {}
    return name, args, kwargs


def bool_param(value, default=None):
    # type: (Union[bool, Text, None], T) -> Union[bool, T]
    """Coverts string param into boolean"""
    if isinstance(value, bool):
        result = value
    else:  # string
        value = "" if value is None else value.lower()
        if value in ("1", "true", "yes", "y", "on"):
            result = True
        elif value in ("0", "false", "no", "n", "off"):
            result = False
        else:
            result = default
    return result


def date_param(value, default=None):
    """Converts string param into datetime.date"""
    if isinstance(value, datetime.date):
        result = value
    else:  # string
        if value:
            result = dateutil_parse(value).date()
        else:
            result = default
    return result


def list_param(
    value,
    default=None,
    delimiter=r"\s+",
    strip=True,
    skip_empty=True,
    replace_map=None,
    cast=None,
):
    """Converts string param into list"""
    if isinstance(value, list):
        result = value
    else:  # string
        if not value:
            result = default or []
        else:
            result = []
            replace_map = replace_map or {}
            cast = cast or text_type
            for item in re.split(delimiter, value):
                if strip:
                    item = item.strip()
                for from_char, to_char in replace_map.items():
                    item = item.replace(from_char, to_char)
                if skip_empty and not item:
                    continue
                result.append(cast(item))
    return result


def flag_param(value, flag, flags):
    """Converts param value into boolean. When explicit value is not provided,
    checks presence of given flag in `flags`.
    List of all flags should be passed as string with values delimited by
    whitespace, ";", "," or "-".
    """
    value = bool_param(value)
    if value is None:
        flags = list_param(flags, delimiter=r"[\s;,\-\|]+")
        value = flag in flags
    return value


def is_accessible_dir(path):
    """Checks if given directory exists
    and current process has rwx permissions to it.
    """
    return os.access(path, os.R_OK & os.W_OK & os.X_OK) and os.path.isdir(path)


def ensure_dir(path):
    """Ensures that given directory exists
    and current process has rwx permissions to it.
    """
    if os.path.exists(path):
        ok = is_accessible_dir(path)
    else:
        try:
            os.makedirs(path)
        except OSError:
            ok = False
        else:
            ok = True
    return ok


def is_cron_process_resumed():
    """Returns True if current process was resumed, False otherwise"""
    return bool_param(os.environ.get("CRON_PROCESS_RESUMED"), default=False)


def is_cron_job_running(cron_job_class, name=None, args=None, kwargs=None):
    """Returns True if cron job is running, False otherwise"""
    from .worker import CronWorker

    worker = CronWorker()
    pid_file = worker.get_pid_file(
        cron_job_class=cron_job_class,
        name=name or cron_job_class.__name__,
        args=args or [],
        kwargs=kwargs or {},
    )
    return worker.pid_file_locked(pid_file, 1)


def spawn(*args, **kwargs):
    """Creates a subprocess that can survive process exit.
    Returns PID of the new process.
    Non-blocking call.
    """
    kwargs["stdout"] = open(os.devnull, "wb")
    kwargs["stderr"] = subprocess.STDOUT
    return subprocess.Popen(args, **kwargs).pid


def execute(*args, **kwargs):
    """Creates a subprocess and waits for it to finish.
    Returns ExecuteResult object which provides return code and output.
    Blocking call.
    """
    kwargs["stdout"] = kwargs.get("stdout", subprocess.PIPE)
    kwargs["stderr"] = kwargs.get("stderr", subprocess.PIPE)
    process = subprocess.Popen(args, **kwargs)
    output, errors = process.communicate()
    status = process.returncode
    return ExecuteResult(args, status, output, errors)


def execute_shell(command, command_args=None, **kwargs):
    """Executes shell command in a subprocess and waits for it to finish.
    Returns ExecuteResult object which provides return code and output.
    Blocking call.
    NOTICE: Should be avoided unless you need shell-specific features.
    """
    if command_args:
        if isinstance(command_args, dict):
            format_args = []
            format_kwargs = {
                key: pipes.quote(text_type(value))
                for key, value in command_args.items()
            }
        else:
            format_args = [
                pipes.quote(text_type(value)) for value in command_args
            ]
            format_kwargs = {}
        command = command.format(*format_args, **format_kwargs)
    kwargs["shell"] = True
    return execute(command, **kwargs)


def execute_pipe(*args_list, **kwargs):
    """Creates subprocess pipe and waits for it to finish.
    Returns ExecuteResult object which provides return code and output
    of the last process.
    Blocking call.
    """
    num_args = len(args_list)
    if num_args < 2:
        raise AssertionError(
            "execute_pipe requires at least 2 positional arguments, "
            "got {}.".format(num_args)
        )
    prev_process = last_process = None
    for args in args_list:
        prev_process = last_process
        popen_kwargs = kwargs.copy()
        popen_kwargs["stdin"] = prev_process.stdout if prev_process else None
        popen_kwargs["stdout"] = subprocess.PIPE
        popen_kwargs["stderr"] = subprocess.PIPE
        last_process = subprocess.Popen(args, **popen_kwargs)
    prev_process.stdout.close()
    output, errors = last_process.communicate()
    status = last_process.returncode
    return ExecuteResult(args_list, status, output, errors)


class ExecuteResult(object):
    """Process execution results"""

    def __init__(self, args, status, output, errors):
        self.args = args
        self.status = status
        self.output = output
        self.errors = errors
        if not self.ok:
            logger.warning(
                "Command `{}` exited with status {}.".format(
                    self.command, self.status
                )
            )

    @cached_property
    def command(self):
        if self.args and not isinstance(self.args[0], (list, tuple)):
            args_list = (self.args,)
        else:
            args_list = self.args
        return " | ".join(
            " ".join(force_text(a) for a in args) for args in args_list
        )

    @property
    def ok(self):
        return self.status == 0

    def __nonzero__(self):
        return self.ok

    @property
    def output_text(self):
        return force_text(self.output)

    @property
    def errors_text(self):
        return force_text(self.errors)

    __bool__ = __nonzero__  # Python 3


class CWD(object):
    """Change working directory utility - decorator and context manager"""

    def __init__(self, new_dir):
        self.new_dir = new_dir
        self.old_dir = None

    def __call__(self, function):
        def _cwd(*args, **kwargs):
            self.change()
            try:
                return function(*args, **kwargs)
            finally:
                self.reset()

        return _cwd

    def __enter__(self):
        self.change()

    def __exit__(self, *args, **kwargs):
        self.reset()

    def change(self):
        """Changes current working directory, remembers previous value"""
        self.old_dir = os.getcwd()
        os.chdir(self.new_dir)

    def reset(self):
        """Restores previous working directory"""
        os.chdir(self.old_dir)


class TabularFormatter(object):
    """Helper to format tabular output for the console"""

    @classmethod
    def format_listing_output(
        cls, items, totals=None, vertical=False, title=None, empty_message=None
    ):
        """Formats listing results (dictionaries) as text"""
        output_lines = []
        if vertical:
            format_line = cls.format_vertical_output_line
        else:
            format_line = cls.format_horizontal_output_line
        if title:
            output_lines.append(title)
        if items:
            for item in items:
                output_lines.append(format_line(item))
            if totals:
                output_lines.append(cls.format_summary_output_line(totals))
        elif empty_message:
            output_lines.append(empty_message)
        return "\n".join(output_lines) + "\n"

    @staticmethod
    def format_horizontal_output_line(data):
        """Converts dictionary values to text separated by tab character"""
        return "\t".join(
            force_text(value)
            for key, value in data.items()
            if not key.startswith("_")
        )

    @staticmethod
    def format_vertical_output_line(data):
        """Converts dictionary to text - `<key>: <value>` pairs separated
        by new line character.
        """
        return "\n".join(
            "{}: {}".format(force_text(key), force_text(value))
            for key, value in data.items()
            if not key.startswith("_")
        )

    @staticmethod
    def format_summary_output_line(data):
        """Converts dictionary to text - `<key>: <value>` pairs separated
        by tab character.
        """
        return "\t".join(
            "{}: {}".format(force_text(key), force_text(value))
            for key, value in data.items()
        )



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django.apps import AppConfig


class CronConfig(AppConfig):
    """Default AppConfig for cronman app."""

    name = "cronman"

    def ready(self):
        """Run app-specific code when Django starts."""
        from cronman import autodiscover

        autodiscover()



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import os

DEBUG = True

SECRET_KEY = "test-secret-key-n8gm2vaj43tqzh3yvbg9x1vj9f3m9mfcq7"

# SQLite:
DATABASES = {
    "default": {"NAME": ":memory:", "ENGINE": "django.db.backends.sqlite3"}
}

INSTALLED_APPS = (
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "django.contrib.admin",
    "cronman",
)

USE_I18N = True

LANGUAGE_CODE = "en"

# ROOT_URLCONF = "cronman.tests.urls"

_log_level = os.environ.get("TEST_CRONMAN_LOG_LEVEL", "INFO")

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {"console": {"class": "logging.StreamHandler"}},
    "loggers": {
        "django": {"handlers": ["console"], "level": _log_level},
        "cronman": {"handlers": ["console"], "level": _log_level},
    },
}



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import datetime

from django.utils import timezone

import mock

from cronman.models import CronTask
from cronman.taxonomies import CronTaskStatus
from cronman.tests.base import (
    BaseCronTestCase,
    create_pid_file,
    override_cron_settings,
    patch_kill,
    patch_ps,
)
from cronman.tests.tools import call_worker


class RunCronTasksTestCase(BaseCronTestCase):
    """Tests for RunCronTasks cron job"""

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.run_cron_tasks.connections.close_all")
    @mock.patch("cronman.cron_jobs.run_cron_tasks.CronSpawner.start_worker")
    @mock.patch("cronman.job.logging.getLogger")
    def test_run_no_workers(self, mock_get_logger, mock_start, mock_close_all):
        """Test for RunCronTasks - no CronTask objects started"""
        self.assertTrue(call_worker("RunCronTasks").ok)
        mock_get_logger.return_value.info.assert_has_calls(
            [mock.call("No CronTasks started.")]
        )
        mock_start.assert_not_called()
        mock_close_all.assert_called_once()

    @override_cron_settings()
    @mock.patch("cronman.cron_jobs.run_cron_tasks.connections.close_all")
    @mock.patch("cronman.cron_jobs.run_cron_tasks.CronSpawner.start_worker")
    @mock.patch("cronman.job.logging.getLogger")
    def test_run_3_workers(self, mock_get_logger, mock_start, mock_close_all):
        """Test for RunCronTasks - 3 CronTask objects started"""
        now = timezone.now()
        # Pending Tasks:
        cron_task_c1 = CronTask.objects.run_now(
            "Sleep", now=now - datetime.timedelta(minutes=6)
        )[0]
        cron_task_c2 = CronTask.objects.run_now(
            "ParamsLockedSleep", now=now - datetime.timedelta(minutes=1)
        )[0]
        cron_task_c3 = CronTask.objects.run_now(
            "ClassLockedSleep", now=now - datetime.timedelta(minutes=1)
        )[0]
        cron_task_c3.mark_as_queued()
        # Future Task:
        CronTask.objects.run_now(
            "ClassLockedSleep",
            params="42",
            now=now + datetime.timedelta(minutes=6),
        )
        # Finished Task:
        CronTask.objects.run_now("IgnoreLockErrorsSleep")[0].mark_as_finished()

        self.assertTrue(call_worker("RunCronTasks").ok)

        mock_get_logger.return_value.info.assert_has_calls(
            [mock.call("Started 3 CronTask(s).")]
        )
        mock_start.assert_has_calls(
            [
                mock.call("Sleep:task_id={}".format(cron_task_c1.pk)),
                mock.call(
                    "ParamsLockedSleep:task_id={}".format(cron_task_c2.pk)
                ),
                mock.call(
                    "ClassLockedSleep:task_id={}".format(cron_task_c3.pk)
                ),
            ]
        )
        mock_close_all.assert_called_once()


class CleanCronTasksTestCase(BaseCronTestCase):
    """Tests for CleanCronTasks cron job"""

    @override_cron_settings()
    @mock.patch("cronman.job.logging.getLogger")
    def test_run_no_workers(self, mock_get_logger):
        """Test for CleanCronTasks - no CronTask object marked as failed."""
        self.assertTrue(call_worker("CleanCronTasks").ok)
        mock_get_logger.return_value.info.assert_has_calls(
            [mock.call("No CronTasks marked as failed.")]
        )

    @override_cron_settings()
    @mock.patch("cronman.job.logging.getLogger")
    def test_run_2_started_1_dead(self, mock_get_logger):
        """Test for CleanCronTasks - 2 CronTask objects started, 1 marked as
        failed.
        """
        now = timezone.now()

        # Started Task - running:
        pid_s1 = 1001
        cron_task_s1 = CronTask.objects.run_now(
            "Sleep", now=now - datetime.timedelta(minutes=6)
        )[0]
        cron_task_s1.mark_as_started(pid_s1)

        # Started Task - dead:
        cron_task_s2 = CronTask.objects.run_now(
            "ParamsLockedSleep", now=now - datetime.timedelta(minutes=1)
        )[0]
        pid_s2 = 1002
        cron_task_s2.mark_as_started(pid_s2)

        # Queued Task:
        cron_task_q1 = CronTask.objects.run_now(
            "ClassLockedSleep", now=now - datetime.timedelta(minutes=1)
        )[0]
        cron_task_q1.mark_as_queued()

        # Waiting Task:
        cron_task_w1 = CronTask.objects.run_now(
            "ClassLockedSleep",
            params="42",
            now=now + datetime.timedelta(minutes=6),
        )[0]

        # Finished Task:
        cron_task_f1 = CronTask.objects.run_now("IgnoreLockErrorsSleep")[0]
        cron_task_f1.mark_as_finished()

        create_pid_file(cron_task_s1.job_spec(), pid_s1)
        with patch_ps(active_pids=[pid_s1]):
            with patch_kill(active_pids=[pid_s1]):
                self.assertTrue(call_worker("CleanCronTasks").ok)

        mock_get_logger.return_value.info.assert_has_calls(
            [mock.call("1 CronTask(s) marked as failed.")]
        )

        # Started Task - running - no status change:
        cron_task_s1.refresh_from_db()
        self.assertEqual(cron_task_s1.status, CronTaskStatus.STARTED)

        # Started Task - dead - marked as failed:
        cron_task_s2.refresh_from_db()
        self.assertEqual(cron_task_s2.status, CronTaskStatus.FAILED)

        # Other statuses - no status change:
        cron_task_q1.refresh_from_db()
        self.assertEqual(cron_task_q1.status, CronTaskStatus.QUEUED)
        cron_task_w1.refresh_from_db()
        self.assertEqual(cron_task_w1.status, CronTaskStatus.WAITING)
        cron_task_f1.refresh_from_db()
        self.assertEqual(cron_task_f1.status, CronTaskStatus.FINISHED)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2018-07-23 03:33
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="CronTask",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("cron_job", models.CharField(max_length=255)),
                ("params", models.TextField(blank=True)),
                (
                    "start_at",
                    models.DateTimeField(
                        blank=True, default=django.utils.timezone.now
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("waiting", "Waiting"),
                            ("queued", "Queued"),
                            ("started", "Started"),
                            ("finished", "Finished"),
                            ("failed", "Failed"),
                        ],
                        default="waiting",
                        max_length=16,
                    ),
                ),
                ("pid", models.PositiveIntegerField(blank=True, null=True)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, null=True),
                ),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("finished_at", models.DateTimeField(blank=True, null=True)),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        )
    ]



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import datetime

import mock

from cronman.exceptions import CronJobNotRegistered
from cronman.scheduler import CronScheduler
from cronman.tests.base import (
    TEMP_FILE,
    BaseCronTestCase,
    override_cron_settings,
)


class CronSchedulerTestCase(BaseCronTestCase):
    """Tests for CronScheduler class"""

    # Main methods

    @override_cron_settings(CRONMAN_JOBS_MODULE=None)
    @mock.patch("cronman.scheduler.scheduler.CronSpawner.start_worker")
    @mock.patch(
        "cronman.scheduler.scheduler.CronScheduler.cron_worker",
        new_callable=mock.PropertyMock,
    )
    def test_run_no_workers(self, mock_cron_worker, mock_start):
        """Test for CronScheduler.run method - no workers started"""
        scheduler = CronScheduler()
        output = scheduler.run()
        self.assertIn("No jobs started.\n", output)
        mock_start.assert_not_called()
        mock_cron_worker.return_value.resume.assert_not_called()

    @override_cron_settings(
        CRONMAN_JOBS_MODULE=None,
        CRONMAN_CRON_SCHEDULER_CRONITOR_ID=None,
    )
    @mock.patch("cronman.monitor.Cronitor.complete")
    @mock.patch("cronman.monitor.Cronitor.fail")
    @mock.patch("cronman.monitor.Cronitor.run")
    def test_run_cronitor_disabled_no_exceptions(
        self, mock_cronitor_run, mock_cronitor_fail, mock_cronitor_complete
    ):
        """Test for CronScheduler.run method - with disabled cronitor calls
        and when there are no exceptions raised that means scheduler will only
        enter in CronScheduler.before_start and CronScheduler.on_success.
        """
        scheduler = CronScheduler()
        scheduler.run()
        mock_cronitor_run.assert_not_called()
        mock_cronitor_fail.assert_not_called()
        mock_cronitor_complete.assert_not_called()

    @override_cron_settings(
        CRONMAN_JOBS_MODULE=None,
        CRONMAN_CRON_SCHEDULER_CRONITOR_ID="<test-cronitor-id>",
    )
    @mock.patch("cronman.monitor.Cronitor.complete")
    @mock.patch("cronman.monitor.Cronitor.fail")
    @mock.patch("cronman.monitor.Cronitor.run")
    def test_run_cronitor_enabled_no_exceptions(
        self, mock_cronitor_run, mock_cronitor_fail, mock_cronitor_complete
    ):
        """Test for CronScheduler.run method - with enabled cronitor calls
        and when there are no exceptions raised that means scheduler will only
        enter in CronScheduler.before_start and CronScheduler.on_success.
        """
        scheduler = CronScheduler()
        scheduler.run()
        mock_cronitor_run.assert_called_once()
        mock_cronitor_fail.assert_not_called()
        mock_cronitor_complete.assert_called_once()

    @override_cron_settings(
        CRONMAN_CRON_SCHEDULER_CRONITOR_ID=None,
        CRONMAN_JOBS_MODULE="cronman.tests.cron_jobs_failing",
    )
    @mock.patch("cronman.monitor.Cronitor.complete")
    @mock.patch("cronman.monitor.Cronitor.fail")
    @mock.patch("cronman.monitor.Cronitor.run")
    def test_run_cronitor_disabled_with_exceptions(
        self, mock_cronitor_run, mock_cronitor_fail, mock_cronitor_complete
    ):
        """Test for CronScheduler.run method - with disabled cronitor calls
        and when there are exceptions raised that means scheduler will only
        enter in CronScheduler.before_start and CronScheduler.on_error.
        """
        with self.assertRaises(CronJobNotRegistered):
            scheduler = CronScheduler()
            scheduler.run()
        mock_cronitor_run.assert_not_called()
        mock_cronitor_fail.assert_not_called()
        mock_cronitor_complete.assert_not_called()

    @override_cron_settings(
        CRONMAN_CRON_SCHEDULER_CRONITOR_ID="<test>",
        CRONMAN_JOBS_MODULE="cronman.tests.cron_jobs_failing",
    )
    @mock.patch("cronman.monitor.Cronitor.complete")
    @mock.patch("cronman.monitor.Cronitor.fail")
    @mock.patch("cronman.monitor.Cronitor.run")
    def test_run_cronitor_enabled_with_exceptions(
        self, mock_cronitor_run, mock_cronitor_fail, mock_cronitor_complete
    ):
        """Test for CronScheduler.run method - with disabled cronitor calls
        and when there are no exceptions raised that means scheduler will only
        enter in CronScheduler.before_start and CronScheduler.on_error.
        """
        with self.assertRaises(CronJobNotRegistered):
            scheduler = CronScheduler()
            scheduler.run()
        mock_cronitor_run.assert_called_once()
        mock_cronitor_fail.assert_called_once()
        mock_cronitor_complete.assert_not_called()

    @override_cron_settings()
    @mock.patch("cronman.scheduler.scheduler.CronSpawner.start_worker")
    @mock.patch(
        "cronman.scheduler.scheduler.CronScheduler.cron_worker",
        new_callable=mock.PropertyMock,
    )
    def test_run_2_workers(self, mock_cron_worker, mock_start):
        """Test for CronScheduler.run method - 2 workers started"""
        scheduler = CronScheduler()
        output = scheduler.run()
        self.assertIn("Started 2 job(s)", output)
        mock_start.assert_has_calls(
            [
                mock.call("Sleep:seconds=1,path={}".format(TEMP_FILE)),
                mock.call("Sleep:seconds=2"),
            ]
        )
        mock_cron_worker.return_value.resume.assert_not_called()

    @override_cron_settings()
    @mock.patch("cronman.scheduler.scheduler.CronSpawner.start_worker")
    @mock.patch(
        "cronman.scheduler.scheduler.CronScheduler.cron_worker",
        new_callable=mock.PropertyMock,
    )
    def test_run_resume(self, mock_cron_worker, mock_start):
        """Test for CronScheduler.run method with resume file"""
        scheduler = CronScheduler()
        scheduler.resume_file.create()
        scheduler.run()
        mock_cron_worker.return_value.resume.assert_called_once()
        self.assertFalse(scheduler.resume_file.exists())  # resume file deleted

    @override_cron_settings()
    @mock.patch("cronman.scheduler.scheduler.CronSpawner.start_worker")
    @mock.patch(
        "cronman.scheduler.scheduler.CronScheduler.cron_worker",
        new_callable=mock.PropertyMock,
    )
    def test_run_disabled(self, mock_cron_worker, mock_start):
        """Test for CronScheduler.run method when lock is acquired"""
        scheduler = CronScheduler()
        scheduler.lock_file.create()
        output = scheduler.run()
        self.assertIn(
            "CronSchedulerLocked: "
            "Scheduler is disabled (lock file exists). "
            'To enable it again, please run "cron_scheduler enable". '
            "Quitting now.\n",
            output,
        )
        mock_start.assert_not_called()
        mock_cron_worker.return_value.resume.assert_not_called()

    @override_cron_settings()
    @mock.patch(
        "cronman.scheduler.scheduler.CronSpawner.start_worker",
        return_value=None,
    )
    @mock.patch(
        "cronman.scheduler.scheduler.CronScheduler.cron_worker",
        new_callable=mock.PropertyMock,
    )
    def test_run_cannot_spawn_process(self, mock_cron_worker, mock_start):
        """Test for CronScheduler.run method - case: cannot spawn process"""
        scheduler = CronScheduler()
        scheduler.slack = mock.MagicMock()
        scheduler.logger = mock.MagicMock()
        output = scheduler.run()
        mock_start.assert_has_calls(
            [
                mock.call("Sleep:seconds=1,path={}".format(TEMP_FILE)),
                mock.call("Sleep:seconds=2"),
            ]
        )
        self.assertIn("No jobs started.", output)
        mock_cron_worker.return_value.resume.assert_not_called()

    @override_cron_settings()
    @mock.patch(
        "cronman.scheduler.scheduler.CronScheduler.cron_worker",
        new_callable=mock.PropertyMock,
    )
    def test_disable(self, mock_cron_worker):
        """Test for CronScheduler.disable method - OK"""
        scheduler = CronScheduler()
        output = scheduler.disable()
        self.assertIn("Scheduler disabled (lock file created).\n", output)
        self.assertTrue(scheduler.lock_file.exists())
        mock_cron_worker.return_value.suspend.assert_not_called()

    @override_cron_settings()
    @mock.patch(
        "cronman.scheduler.scheduler.CronScheduler.cron_worker",
        new_callable=mock.PropertyMock,
    )
    def test_disable_with_workers(self, mock_cron_worker):
        """Test for CronScheduler.disable method with workers option - OK"""
        scheduler = CronScheduler()
        output = scheduler.disable(workers=True)
        self.assertIn(
            "Scheduler disabled (lock file created, workers suspended).\n",
            output,
        )
        self.assertTrue(scheduler.lock_file.exists())
        mock_cron_worker.return_value.suspend.assert_called_once()

    @override_cron_settings()
    @mock.patch(
        "cronman.scheduler.scheduler.CronScheduler.cron_worker",
        new_callable=mock.PropertyMock,
    )
    def test_disable_already_disabled(self, mock_cron_worker):
        """Test for CronScheduler.disable method - FAIL, already disabled"""
        scheduler = CronScheduler()
        scheduler.lock_file.create()
        output = scheduler.disable()
        self.assertIn(
            "CronSchedulerLocked: "
            "Scheduler is already disabled (lock file exists).",
            output,
        )
        self.assertTrue(scheduler.lock_file.exists())
        mock_cron_worker.return_value.suspend.assert_not_called()

    @override_cron_settings()
    def test_enable(self):
        """Test for CronScheduler.enable method - OK"""
        scheduler = CronScheduler()
        scheduler.lock_file.create()
        output = scheduler.enable()
        self.assertIn("Scheduler enabled (lock file deleted).\n", output)
        self.assertFalse(scheduler.lock_file.exists())
        self.assertFalse(scheduler.resume_file.exists())

    @override_cron_settings()
    def test_enable_with_workers(self):
        """Test for CronScheduler.enable method with workers option - OK"""
        scheduler = CronScheduler()
        scheduler.lock_file.create()
        output = scheduler.enable(workers=True)
        self.assertIn(
            "Scheduler enabled (resume file created, lock file deleted).\n",
            output,
        )
        self.assertFalse(scheduler.lock_file.exists())
        self.assertTrue(scheduler.resume_file.exists())

    @override_cron_settings()
    def test_enable_already_enabled(self):
        """Test for CronScheduler.enable method - FAIL, already enabled"""
        scheduler = CronScheduler()
        output = scheduler.enable()
        self.assertIn(
            "CronSchedulerUnlocked: "
            "Scheduler is already enabled (lock file does not exist).",
            output,
        )
        self.assertFalse(scheduler.lock_file.exists())
        self.assertFalse(scheduler.resume_file.exists())

    # Helpers

    @override_cron_settings()
    def test_get_datetime_range(self):
        """Test for CronScheduler.get_datetime_range method"""
        now = datetime.datetime(2017, 5, 13, 12, 1, 3, 321)
        self.assertEqual(
            CronScheduler(now=now).get_datetime_range(),
            (
                datetime.datetime(2017, 5, 13, 12, 0, 59),
                datetime.datetime(2017, 5, 13, 12, 2, 59),
            ),
        )

    @override_cron_settings()
    def test_cron_jobs(self):
        """Test that cron_jobs property is correct (CRON_JOBS property of
        settings.CRONMAN_JOBS_MODULE).
        """
        scheduler = CronScheduler()
        from cronman.tests.cron_jobs import CRON_JOBS

        self.assertEqual(scheduler.cron_jobs, CRON_JOBS)

    @override_cron_settings(CRONMAN_JOBS_MODULE=None)
    def test_cron_jobs_empty(self):
        """Test that cron_jobs property is empty tuple when CRONMAN_JOBS_MODULE
        is not set.
        """
        scheduler = CronScheduler()
        self.assertEqual(scheduler.cron_jobs, ())

    @override_cron_settings(CRONMAN_JOBS_MODULE="module.does.not.exist")
    def test_cron_jobs_import_error(self):
        """Test that cron_jobs property raises ImportError when
        CRONMAN_JOBS_MODULE is not a valid module."""
        scheduler = CronScheduler()
        with self.assertRaises(ImportError):
            self.assertEqual(scheduler.cron_jobs, ())

    @override_cron_settings()
    def test_get_jobs(self):
        """Test for CronScheduler.get_jobs method"""
        now = datetime.datetime(2017, 5, 13, 12, 1, 3, 321)
        self.assertEqual(
            CronScheduler(now=now).get_jobs(),
            [
                ("*/2 * * * *", "Sleep:seconds=1,path={}".format(TEMP_FILE)),
                ("*/2 * * * *", "Sleep:seconds=2"),
            ],
        )



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from cronman.cron_jobs.clean_cron_tasks import CleanCronTasks
from cronman.cron_jobs.run_cron_tasks import RunCronTasks
from cronman.cron_jobs.sleep import (
    ClassLockedSleep,
    IdleIOSleep,
    IgnoreLockErrorsSleep,
    LowCPUSleep,
    LowestCPUIOSleep,
    ParamsLockedSleep,
    PersistentSleep,
    PersistentSleep2,
    SlackNotifyDoneSleep,
    Sleep,
)
from cronman.job import cron_job_registry

cron_job_registry.register(RunCronTasks)
cron_job_registry.register(CleanCronTasks)
cron_job_registry.register(Sleep)
cron_job_registry.register(ClassLockedSleep)
cron_job_registry.register(ParamsLockedSleep)
cron_job_registry.register(IgnoreLockErrorsSleep)
cron_job_registry.register(SlackNotifyDoneSleep)
cron_job_registry.register(LowCPUSleep)
cron_job_registry.register(LowestCPUIOSleep)
cron_job_registry.register(IdleIOSleep)
cron_job_registry.register(PersistentSleep)
cron_job_registry.register(PersistentSleep2)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django.core.management import CommandError

# General errors:


class MissingDependency(ImportError):
    """Exception raised when missing optional dependency is accessed."""


# CronJobRegistry errors:


class CronJobRegistryError(KeyError):
    """Error related to CronJobRegistry"""


class CronJobAlreadyRegistered(CronJobRegistryError):
    """CronJob class of given name is already registered"""


class CronJobNotRegistered(CronJobRegistryError):
    """CronJob class of given name is not registered"""


# CronScheduler errors:


class CronSchedulerError(CommandError):
    """General CronScheduler error"""


class CronSchedulerLocked(CronSchedulerError):
    """CronScheduler is locked"""


class CronSchedulerUnlocked(CronSchedulerError):
    """CronScheduler is unlocked"""


class CronSchedulerNoJobs(CronSchedulerError):
    """CronScheduler no jobs to start"""


# CronWorker errors:


class CronWorkerError(CommandError):
    """General CronScheduler error"""


class PIDAccessError(CronWorkerError):
    """CronWorker has no access to process by PID"""


class CronTaskInvalidStatus(CronWorkerError):
    """CronWorker received CronTask with improper status"""


class CronWorkerLocked(CronWorkerError):
    """CronWorker cannot start due to active lock"""


class CronWorkerInvalidParams(CronWorkerError):
    """CronWorker received invalid arguments"""



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

from django.utils.module_loading import autodiscover_modules

from cronman.job import cron_job_registry

default_app_config = "cronman.apps.CronConfig"


def autodiscover():
    """Discovers `cron_jobs` modules in other apps,
    fills out `cron_job_registry`
    """
    autodiscover_modules("cron_jobs", register_to=cron_job_registry)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

from __future__ import unicode_literals

import hashlib
import os

from django.utils.encoding import force_bytes, force_text
from django.utils.functional import cached_property

from cronman.spawner import CronSpawner
from cronman.worker.process_manager import ProcessManager


class BaseCronWorkerFile(object):
    """Stats file kept by Cron Worker - base class"""

    EXTENSION = None  # must be set in subclass

    def __init__(self, data_dir, name):
        self.name = name
        self.data_dir = data_dir
        self.path = os.path.join(
            self.data_dir, "{}{}".format(self.name, self.EXTENSION)
        )

    @classmethod
    def get_file_name(cls, name, args=None, kwargs=None, random=False):
        """Generates file name from CronJob's name, args and kwargs"""
        parts = [name]
        if args or kwargs:
            params_bytes_repr = force_bytes(repr((args, kwargs)))
            parts.append(
                hashlib.md5(params_bytes_repr).hexdigest()[:10]  # nosec
            )
        if random:
            parts.append(hashlib.md5(os.urandom(16)).hexdigest()[:10])  # nosec
        return "_".join(parts)

    @classmethod
    def all(cls, data_dir, name=None, args=None, kwargs=None):
        """Iterates over files in given directory"""
        for file_name in os.listdir(data_dir):
            base_name, ext = os.path.splitext(file_name)
            if ext == cls.EXTENSION:
                if name:  # Filter files by specific job:
                    base_name_begin = cls.get_file_name(name, args, kwargs)
                    if base_name.startswith(base_name_begin):
                        yield cls(data_dir, base_name)
                else:  # No filters:
                    yield cls(data_dir, base_name)

    def write_content(self, content):
        """Populates this file with new content"""
        with open(self.path, "wb") as file_:
            file_.write(force_bytes(content))
            file_.flush()
            # NOTE:
            # "flush() does not necessarily write the file’s data to disk.
            # Use flush() followed by os.fsync() to ensure this behavior."
            # (from Python 2.7 docs)
            os.fsync(file_.fileno())

    def read_content(self):
        """Reads contents of this file"""
        try:
            with open(self.path, "rb") as file_:
                content = force_text(file_.read())
        except IOError:  # File deleted already
            content = None
        return content

    def delete(self):
        """Deletes this file"""
        try:
            os.unlink(self.path)
        except OSError:
            pass  # File deleted already

    def exists(self):
        """Checks if this file exists"""
        return os.path.exists(self.path)


class CronWorkerPIDFile(BaseCronWorkerFile):
    """PID file (and lock) for Cron Worker"""

    EXTENSION = ".pid"

    @classmethod
    def by_pid(cls, data_dir, pid):
        """Retrieves PID file from given directory by PID value"""
        for candidate_pid_file in cls.all(data_dir):
            if candidate_pid_file.pid == pid:
                pid_file = candidate_pid_file
                break
        else:
            pid_file = None
        return pid_file

    def create(self):
        """Creates the PID file"""
        pid = os.getpid()
        self.write_content(pid)
        self.__dict__["pid"] = pid  # Bypass `pid` property

    @cached_property
    def pid(self):
        """PID extracted from this file"""
        try:
            pid = int(self.read_content())
        except (ValueError, TypeError):  # PID file deleted / truncated
            pid = None
        return pid

    @cached_property
    def process(self):
        """Process Manager instance for this PID file"""
        return ProcessManager(self.pid)  # ProcessManager can handle `None`

    def exists_with_alive_process(self):
        """Checks if PID file exists and corresponding process is alive.
        Removes file if process is dead.
        """
        if self.exists():
            if not self.pid or self.process.alive() is False:
                # PID file deleted/truncated or process is dead:
                self.delete()
                result = False
            else:  # Process is alive or we don't have access:
                result = True
        else:
            result = False
        return result

    @cached_property
    def job_spec_file(self):
        """JobSpec file associated with this file"""
        return CronWorkerJobSpecFile(self.data_dir, self.name)


class CronWorkerJobSpecFile(BaseCronWorkerFile):
    """JobSpec file for Cron Worker's running jobs with resume feature"""

    EXTENSION = ".jobspec"

    @classmethod
    def by_pid(cls, data_dir, pid):
        """Retrieves JobSpec file from given directory by PID value"""
        job_spec_file = None
        pid_file = CronWorkerPIDFile.by_pid(data_dir, pid)
        if pid_file:
            job_spec_file = cls(data_dir, pid_file.name)
            if not job_spec_file.exists():
                job_spec_file = None
        return job_spec_file

    def create(self, job_spec):
        """Creates the JobSpec file"""
        self.write_content(job_spec)
        self.__dict__["job_spec"] = job_spec  # Bypass `job_spec` property

    def resume(self):
        """Starts a worker process for job spec stored in this file"""
        job_spec = self.job_spec
        self.delete()  # file is deleted before spawning new worker
        if job_spec:
            pid = self.cron_spawner.start_worker(job_spec)
        else:
            pid = None
        return pid

    @cached_property
    def job_spec(self):
        """JobSpec extracted from this file"""
        return self.read_content() or None

    @cached_property
    def pid_file(self):
        """PID file associated with this file"""
        return CronWorkerPIDFile(self.data_dir, self.name)

    @cached_property
    def cron_spawner(self):
        """Cron Spawner instance"""
        return CronSpawner(
            data_dir=self.data_dir, extra_env={"CRON_PROCESS_RESUMED": "1"}
        )
